from flask import current_app

from keg_elements import sentry, testing


class TestFilterConfig:
    @classmethod
    def setup_class(cls):
        cls.sentry = sentry.SentryClient(dsn='https://test:test@sentry.io/test')

    def check_config_filtered_in_stack(self, stack, *keys):
        frames = stack['frames']
        instance_count = 0
        for frame in frames:
            scope = frame['vars']
            for key in keys:
                if key in scope:
                    instance_count += 1
                    assert scope[key] == '{ ... }'
        return instance_count

    def test_filtered_from_stack(self):
        def do_raise():
            config = current_app.config  # noqa
            try:
                foo = current_app.config  # noqa
                raise ValueError('Some error')
            except ValueError:
                self.sentry.captureException()

        with testing.SentryCapture() as cap:
            do_raise()
        assert len(cap.reports) == 1
        stack = cap.reports[0]['exception']['values'][0]['stacktrace']
        assert self.check_config_filtered_in_stack(stack, 'config', 'foo') == 2

    def test_filtering_serializer(self):
        serializer = sentry.ConfigTypeFilterSerializer(None)

        assert serializer.can(current_app.config)
        assert not serializer.can({})

        assert serializer.serialize(current_app.config) == '{ ... }'



import flask
import raven
from raven.utils.serializer.base import Serializer
from raven.utils.serializer.manager import transform, SerializationManager


class ConfigTypeFilterSerializer(Serializer):
    """
    Sentry serializer for the Config type that will hide the contents
    """
    TYPES = (
        flask.config.Config,
    )

    def can(self, value):
        return isinstance(value, self.TYPES)

    def serialize(self, value, **kwargs):
        return '{ ... }'


class FilteringManager(SerializationManager):
    """
    This is used by sentry to find the available serializers to use when building a report.
    Typically we could register a serializer without subclassing this but because the built-in
    serializers take precedent to custom ones we could register and the Config type is a subclass
    of dict, we must insert our serializer before the built ins to force sentry to prefer it over
    the dict serializer.
    """

    def __init__(self, extra_serializers=tuple()):
        self.extra_serializers = extra_serializers
        super(FilteringManager, self).__init__()

    @property
    def serializers(self):
        for es in self.extra_serializers:
            yield es
        for s in SerializationManager.serializers.fget(self):
            yield s


class SentryClient(raven.base.Client):
    """
    A custom sentry client type that includes sanitization for the config type and features to ease
    testing of the reports that would be sent to the sentry API
    """

    # The following attributes are used for testing to capture reports that would be sent.
    # When __log_reports__ is True, any data that would be sent to sentry is stored in the
    # __report_log__ list.  See testing.SentryCapture for a helpful context manager that makes
    # use of these attributes
    __log_reports__ = False
    __report_log__ = []

    def __init__(self, *args, **kwargs):
        self.extra_serializers = kwargs.pop('extra_serializers', [ConfigTypeFilterSerializer])
        super(SentryClient, self).__init__(*args, **kwargs)

    def transform(self, data):
        # Called by sentry to convert data into a JSON friendly format that can be sent to their API
        return transform(
            data,
            list_max_length=self.list_max_length,
            string_max_length=self.string_max_length,
            manager=FilteringManager(self.extra_serializers)
        )

    def send(self, auth_header=None, **data):
        from keg import current_app

        if self.__log_reports__:
            # Report logging is enabled. Store the report in __report_log__
            self.__report_log__.append(data)
            return

        if flask.has_app_context() and current_app.config.get('TESTING'):
            # We are in a test. Don't pass on the report to Sentry.
            return

        # If we are not testing or we cannot determine that because we are not within an
        # app context, pass the report through
        return super(SentryClient, self).send(auth_header=auth_header, **data)  # pragma: no cover



import keg_elements.testing as testing
from kegel_app.model import entities


def test_dont_care():
    assert testing.DontCare() == True  # noqa
    assert (testing.DontCare() == True) is True  # noqa
    assert (testing.DontCare() == False) is True  # noqa
    assert (testing.DontCare() == object) is True  # noqa
    assert (testing.DontCare() != object) is False  # noqa
    assert (testing.DontCare()() == object) is True  # noqa
    assert (testing.DontCare()['thing'] == object) is True  # noqa
    assert (testing.DontCare()[1] == object) is True  # noqa


class TestEntityBaseThing(testing.EntityBase):
    entity_cls = entities.Thing

    column_checks = [
        testing.ColumnCheck('name'),
        testing.ColumnCheck('color', required=False),
        testing.ColumnCheck('scale_check', required=False),
        testing.ColumnCheck('float_check', required=False),
        testing.ColumnCheck('units', required=False),
    ]


class TestEntityBaseAncillaryA(testing.EntityBase):
    entity_cls = entities.AncillaryA

    column_checks = [
        testing.ColumnCheck('thing_id', fk='things.id')
    ]


class TestEntityBaseAncillaryB(testing.EntityBase):
    entity_cls = entities.AncillaryB

    column_checks = [
        testing.ColumnCheck('thing_id', fk='things.id')
    ]


class TestEntityBaseUsesBothFkSet(testing.EntityBase):
    entity_cls = entities.UsesBoth

    column_checks = [
        testing.ColumnCheck('thing_id', fk={'ancillary_as.thing_id', 'ancillary_bs.thing_id'}),
        testing.ColumnCheck('ancillary_a_id', fk='ancillary_as.id'),
        testing.ColumnCheck('ancillary_b_id', fk='ancillary_bs.id'),
    ]


class TestEntityBaseUsesBothFkList(testing.EntityBase):
    entity_cls = entities.UsesBoth

    column_checks = [
        testing.ColumnCheck('thing_id', fk=['ancillary_as.thing_id', 'ancillary_bs.thing_id']),
        testing.ColumnCheck('ancillary_a_id', fk='ancillary_as.id'),
        testing.ColumnCheck('ancillary_b_id', fk='ancillary_bs.id'),
    ]


class TestEntityBaseUsesBothFkString(testing.EntityBase):
    entity_cls = entities.UsesBoth

    column_checks = [
        testing.ColumnCheck('thing_id', fk='ancillary_as.thing_id, ancillary_bs.thing_id'),
        testing.ColumnCheck('ancillary_a_id', fk='ancillary_as.id'),
        testing.ColumnCheck('ancillary_b_id', fk='ancillary_bs.id'),
    ]



import codecs
import enum
import random

import sqlalchemy as sa

from keg.db import db
import keg_elements.db.mixins as mixins
import keg_elements.db.columns as columns


many_things_mapper = db.Table(
    'many_things_mapper',
    sa.Column('thing_id', sa.Integer, sa.ForeignKey('things.id')),
    sa.Column('many_id', sa.Integer, sa.ForeignKey('many_things.id'))
)


class Units(enum.Enum):
    feet = 'ft'
    inches = 'in'
    meters = 'm'


class Thing(db.Model, mixins.DefaultMixin):
    __tablename__ = 'things'

    name = db.Column(db.Unicode(50), nullable=False)
    color = db.Column(db.Unicode)
    scale_check = db.Column(db.Numeric(8, 4))
    float_check = db.Column(db.Float)
    units = db.Column(sa.Enum(Units, name='enum_units'))

    @sa.ext.hybrid.hybrid_property
    def name_and_color(self):
        return '{}-{}'.format(self.name, self.color)

    @name_and_color.expression
    def name_and_color(cls):
        return cls.name + sa.sql.literal('-') + cls.color


class RelatedThing(db.Model, mixins.DefaultMixin):
    __tablename__ = 'related_things'

    name = db.Column(db.Unicode(50), nullable=False)
    is_enabled = db.Column(db.Boolean, nullable=False, default=False,
                           server_default=sa.text('FALSE'))

    thing_id = sa.Column(sa.Integer, sa.ForeignKey(Thing.id), nullable=False)
    thing = sa.orm.relationship(lambda: Thing, backref=sa.orm.backref(
        'related_things',
        cascade='all, delete-orphan'
    ))

    @classmethod
    def testing_create(cls, **kwargs):
        kwargs['thing'] = kwargs.get('thing') or Thing.testing_create()
        return super(RelatedThing, cls).testing_create(**kwargs)


class ManyToManyThing(db.Model, mixins.DefaultMixin):
    __tablename__ = 'many_things'

    name = db.Column(db.Unicode(50), nullable=False)

    things = sa.orm.relationship(lambda: Thing,
                                 secondary=many_things_mapper,
                                 backref='manys')


class ThingWithRequiredBoolean(db.Model, mixins.DefaultMixin):
    __tablename__ = 'required_boolean_table'

    nullable_boolean = db.Column(db.Boolean, nullable=True)
    required_boolean = db.Column(db.Boolean, nullable=False)
    required_boolean_with_default = db.Column(db.Boolean, nullable=False, default=False)
    required_boolean_with_server_default = db.Column(db.Boolean, nullable=False,
                                                     server_default='false')


class MultiplePrimaryKeys(db.Model, mixins.DefaultMixin):
    __tablename__ = 'multikey_table'

    other_pk = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.Unicode(50))


def super_secure_encrypt(data, key):
    return codecs.encode(data, 'rot_13').encode()


def super_secure_decrypt(data, key):
    return codecs.encode(data.decode(), 'rot_13')


class ColumnTester(db.Model, mixins.DefaultMixin):
    __tablename__ = 'column_tester'

    time_zone = db.Column(columns.TimeZoneType(length=100))
    encrypted1 = db.Column(columns.EncryptedUnicode(key=lambda: b'a' * 32))
    encrypted2 = db.Column(columns.EncryptedUnicode(key=b'b' * 32))
    encrypted3 = db.Column(columns.EncryptedUnicode(
        key=b'c' * 32,
        encrypt=super_secure_encrypt,
        decrypt=super_secure_decrypt
    ))


class AncillaryA(mixins.DefaultMixin, db.Model):
    __tablename__ = 'ancillary_as'

    thing_id = sa.Column(sa.Integer, sa.ForeignKey(Thing.id), nullable=False)

    @classmethod
    def testing_create(cls, **kwargs):
        if 'thing_id' not in kwargs:
            kwargs['thing_id'] = Thing.testing_create().id

        return super(AncillaryA, cls).testing_create(**kwargs)


class AncillaryB(mixins.DefaultMixin, db.Model):
    __tablename__ = 'ancillary_bs'

    thing_id = sa.Column(sa.Integer, sa.ForeignKey(Thing.id), nullable=False)

    @classmethod
    def testing_create(cls, **kwargs):
        if 'thing_id' not in kwargs:
            kwargs['thing_id'] = Thing.testing_create().id

        return super(AncillaryB, cls).testing_create(**kwargs)


class UsesBoth(mixins.DefaultMixin, db.Model):
    __tablename__ = 'uses_boths'
    __table_args__ = (
        sa.ForeignKeyConstraint(
            ('thing_id', 'ancillary_a_id'),
            (AncillaryA.thing_id, AncillaryA.id)
        ),
        sa.ForeignKeyConstraint(
            ('thing_id', 'ancillary_b_id'),
            (AncillaryB.thing_id, AncillaryB.id)
        ),
    )

    thing_id = sa.Column(sa.Integer, nullable=False)
    ancillary_a_id = sa.Column(sa.Integer, nullable=False)
    ancillary_b_id = sa.Column(sa.Integer, nullable=False)

    @classmethod
    def testing_create(cls, **kwargs):
        if 'thing_id' not in kwargs:
            kwargs['thing_id'] = Thing.testing_create().id

        thing_id = kwargs['thing_id']

        if 'ancillary_a_id' not in kwargs:
            kwargs['ancillary_a_id'] = AncillaryA.testing_create(thing_id=thing_id).id

        if 'ancillary_b_id' not in kwargs:
            kwargs['ancillary_b_id'] = AncillaryB.testing_create(thing_id=thing_id).id

        return super(UsesBoth, cls).testing_create(**kwargs)


class ConstraintTester(mixins.DefaultMixin, db.Model):
    unique1 = sa.Column(sa.Integer, nullable=False, unique=True)
    unique2 = sa.Column(sa.Integer, nullable=False)

    check = sa.Column(sa.Integer, nullable=False)

    __tablename__ = 'constraint_tester'
    __table_args__ = (
        sa.UniqueConstraint(unique2, name='uq_constraint_tester_unique2'),
        sa.CheckConstraint(check <= 100, name='ck_constraint_tester_check'),
    )

    @classmethod
    def testing_create(cls, **kwargs):
        kwargs.setdefault('check', random.randint(0, 100))
        return super(ConstraintTester, cls).testing_create(**kwargs)



import contextlib
import math
import random
from decimal import Decimal

import pytest
import wrapt
import sqlalchemy as sa
from sqlalchemy.sql import expression
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.types import DateTime

from blazeutils.strings import randchars

from keg.db import db
from keg_elements.extensions import lazy_gettext as _


class utcnow(expression.FunctionElement):
    type = DateTime()


@compiles(utcnow, 'postgresql')
def _pg_utcnow(element, compiler, **kw):
    return "TIMEZONE('utc', CURRENT_TIMESTAMP)"


@compiles(utcnow, 'mssql')
def _ms_utcnow(element, compiler, **kw):
    return "GETUTCDATE()"


@compiles(utcnow, 'sqlite')
def _sqlite_utcnow(element, compiler, **kw):
    return "CURRENT_TIMESTAMP"


@wrapt.decorator
def no_autoflush(wrapped, instance, args, kwargs):
    autoflush = db.session.autoflush
    db.session.autoflush = False
    try:
        return wrapped(*args, **kwargs)
    finally:
        db.session.autoflush = autoflush


def random_numeric(column):
    fractional_digits = column.type.scale
    whole_digits = column.type.precision - fractional_digits

    # only use about half the digits to make arithmetic done with this less likely to overflow
    max_whole = 10 ** math.ceil(whole_digits / 2.0) - 1

    whole = random.randint(-max_whole, max_whole)

    fractional = Decimal(random.randint(0, 10 ** fractional_digits - 1)) / 10 ** fractional_digits
    return fractional + whole


def validate_unique_exc(exc, constraint_name=None):
    return _validate_unique_msg(db.engine.dialect.name, str(exc), constraint_name)


def _validate_unique_msg(dialect, msg, constraint_name=None):
    """
        Does the heavy lifting for validate_unique_exception().

        Broken out separately for easier unit testing.  This function takes string args.
    """
    if constraint_name is not None and dialect != 'sqlite' and constraint_name not in msg:
        return False

    if dialect == 'postgresql':
        return 'duplicate key value violates unique constraint' in msg
    elif dialect == 'mssql':
        return 'Cannot insert duplicate key' in msg
    elif dialect == 'sqlite':
        return 'UNIQUE constraint failed' in msg
    else:
        raise ValueError('is_unique_exc() does not yet support dialect: %s' % dialect)


@contextlib.contextmanager
def raises_unique_exc(constraint_name):
    with pytest.raises(sa.exc.IntegrityError) as exc:
        yield
    assert validate_unique_exc(exc.value, constraint_name)


@contextlib.contextmanager
def raises_check_exc(contraint_name):
    with pytest.raises(sa.exc.IntegrityError) as exc:
        yield
    assert _is_check_const(db.engine.dialect.name, str(exc.value), contraint_name)


def _is_check_const(dialect, msg, constraint_name):
    if dialect == 'mssql':
        return 'conflicted with the CHECK constraint' in msg and constraint_name in msg
    elif dialect == 'sqlite':
        return (
            'CHECK constraint {} failed'.format(constraint_name) in msg
            or 'CHECK constraint failed' in msg
        )
    elif dialect == 'postgresql':
        return 'violates check constraint' in msg and constraint_name in msg
    else:
        raise ValueError('is_constraint_exc() does not yet support dialect: %s' % dialect)


def randemail(length, randomizer=randchars):
    """Generate a random email address at the given length.
    :param length: must be at least 7 or the function will throw a ValueError.
    :param randomizer: is a function for generating random characters. It must have an identical
                      interface to `randchars`. The default function is `randchars`.
    """
    if length < 7:
        raise ValueError(_('length must be at least 7'))

    half = (length - 2 - 3) / 2.0  # 2 characters for @ and . and 3 for TLD
    return (randomizer(int(math.floor(half)), 'alphanumeric')
            + '@' + randomizer(int(math.ceil(half)), 'alphanumeric')
            + '.' + randomizer(3, 'alpha'))


def session_commit():
    try:
        db.session.commit()
    except Exception:
        db.session.rollback()
        raise


def session_flush():
    try:
        db.session.flush()
    except Exception:
        db.session.rollback()
        raise


class CollectionUpdater(object):

    @no_autoflush
    def __init__(self, entity, attr_name, data):
        self.entity = entity
        self.attr_name = attr_name
        self.data = data
        self.collection = getattr(entity, attr_name)
        self.keep_children = set()

        ent_cls = entity.__class__
        queryable_attr = getattr(ent_cls, attr_name)
        # Unless they sent in the wrong type, rel_prop should be a
        # sqlalchemy.orm.properties.RelationshipProperty instance
        rel_prop = queryable_attr.property
        # now that we have the property, go through the mapper to get to the child entity class
        self.child_cls = rel_prop.mapper.class_

    @no_autoflush
    def find_child(self, data):
        """Find the child record associated with the related object"""
        child_entity_keys = self.child_cls.primary_keys()
        supplied_keys = set(data.get(column.key)
                            for column in child_entity_keys)

        if None in supplied_keys:
            return None

        for child in self.collection:
            record_keys = set(getattr(child, col.key)
                              for col in child_entity_keys)
            if record_keys == supplied_keys:
                return child
        return None

    @no_autoflush
    def update(self):
        """Update the objects associated with the entity"""
        for record in self.data:
            child = self.find_child(record)
            state = sa.inspect(child) if child else None

            if state and state.persistent:
                child.edit(_commit=False, **record)
            else:
                child = self.child_cls.add(_commit=False, **record)
                self.collection.append(child)

            self.keep_children.add(child)
        self.remove_unmodified()

    def remove_unmodified(self):
        remove_children = [child for child in self.collection if child not in self.keep_children]
        for child in remove_children:
            self.collection.remove(child)



from collections import namedtuple

from keg import current_app
import six
import sqlalchemy as sa
from sqlalchemy_utils import ArrowType

from keg_elements.sentry import SentryClient
from .db.utils import validate_unique_exc
from .db.mixins import DefaultColsMixin

ColumnCheck = namedtuple('ColumnCheck', 'name, required, fk, unique, timestamp')
ColumnCheck.__new__.__defaults__ = (True, None, None, None)


class DontCare(object):  # pragma: no cover
    """A placeholder object that can conform to almost anything and is always
    equal to anything it is compared against.

    Examples:

        assert [1, 2, 3] == [1, DontCare(), 3]
    """
    __eq__ = lambda *_: True
    __ne__ = lambda *_: False
    __repr__ = lambda *_: '_'

    __getattr__ = lambda *_: DontCare()
    __getitem__ = lambda *_: DontCare()
    __call__ = lambda *a, **kw: DontCare()


class EntityBase(object):
    entity_cls = None
    column_checks = None
    timestamp_cols = ('created_utc', 'updated_utc')
    # Set to: 'setup_class', 'setup', or None
    delete_all_on = 'setup_class'

    @classmethod
    def setup_class(cls):
        assert current_app.db_enabled, 'app.db_enabled is False'
        assert cls.entity_cls, 'The entity_cls attribute must be set.'
        assert cls.column_checks, 'column_checks should be defined for this entity'
        if cls.timestamp_cols:
            if cls.column_checks is None:
                cls.column_checks = []
            createdts_col_name, updatedts_col_name = cls.timestamp_cols
            cls.column_checks.append(ColumnCheck(createdts_col_name, timestamp='create'))
            cls.column_checks.append(ColumnCheck(updatedts_col_name, timestamp='update'))
        if cls.delete_all_on:
            cls.ent_delete_all()

    @classmethod
    def column_check_generator(cls):
        for col_check in cls.column_checks or []:
            yield col_check
        if cls.timestamp_cols:
            createdts_col_name, updatedts_col_name = cls.timestamp_cols
            yield ColumnCheck(createdts_col_name, timestamp='create')
            yield ColumnCheck(updatedts_col_name, timestamp='update')

    @classmethod
    def ent_delete_all(cls):
        if hasattr(cls.entity_cls, 'delete_cascaded'):
            cls.entity_cls.delete_cascaded()
        else:
            cls.entity_cls.delete_all()

    def setup(self):
        if self.delete_all_on == 'setup':
            self.ent_delete_all()

    def test_add(self):
        self.ent_delete_all()
        o = self.entity_cls.testing_create()
        assert self.entity_cls.query.count() == 1
        if hasattr(self.entity_cls, 'id'):
            assert o.id

    def test_column_null(self):
        for col_check in self.column_check_generator():
            col = getattr(self.entity_cls, col_check.name)
            assert col.nullable != col_check.required, \
                'Expected column "{}" to match null requirement'.format(col.name)

    def test_column_unique(self):
        for col_check in self.column_check_generator():
            col = getattr(self.entity_cls, col_check.name)
            assert col.unique == col_check.unique, 'Expected column "{}" to have unique={}'.format(
                col.name, col_check.unique)

    def test_column_fk(self):
        for col_check in self.column_check_generator():
            col = getattr(self.entity_cls, col_check.name)
            fk_count = len(col.foreign_keys)
            if col_check.fk:
                # normalize `fk` into a set
                if isinstance(col_check.fk, six.string_types):
                    # 'foo.bar' => {'foo.bar'}
                    # 'foo.bar,baz.qux' => {'foo.bar', 'baz.qux'}
                    # 'foo.bar, baz.qux' => {'foo.bar', 'baz.qux'}
                    fk_set = {partial.strip() for partial in col_check.fk.split(',')}

                elif isinstance(col_check.fk, set):
                    fk_set = col_check.fk

                else:
                    # ['foo.bar', 'baz.qux'] => {'foo.bar', 'baz.qux'}
                    fk_set = set(col_check.fk)

                assert fk_count == len(fk_set), 'FK count does not match for {}'.format(col.name)
                assert fk_set == {fk._get_colspec() for fk in col.foreign_keys}, \
                    'FK set does not match for {}'.format(col.name)

            else:
                assert not fk_count, \
                    'Didn\'t expect column "{}" to have a foreign key'.format(col.name)

    def test_column_timestamp(self):
        for col_check in self.column_check_generator():
            if not col_check.timestamp:
                continue
            col = getattr(self.entity_cls, col_check.name)
            if not isinstance(col.type, (ArrowType, sa.DateTime)):
                raise AssertionError(
                    'Column "{}" is not a recognized timestamp type.'.format(col.name))
            assert col.default
            if col_check.timestamp == 'update':
                assert col.onupdate, 'Column "{}" should have onupdate set'.format(col.name)
            assert col.server_default, 'Column "{}" should have server_default set'.format(col.name)

    def test_all_columns_are_constraint_tested(self):
        """Checks that all fields declared on entity are in the constraint tests"""

        expected_columns = [col.name for col in self.entity_cls.__table__.columns]
        constraint_columns = [col[0] for col in self.column_checks]
        inherited_columns = []

        # Even if the entity class is based (eventually) on DefaultColsMixin, polymorphic subclasses
        #   will not have the default cols, as the parent entity alone will have them
        if isinstance(self.entity_cls(), DefaultColsMixin) and not \
                getattr(self.entity_cls, '__mapper_args__', {}).get('polymorphic_identity', None):
            # Include columns from common import base classes
            for field in vars(DefaultColsMixin):

                # Ignore the usual __class__, __dict__, blah, blah blah
                if field.startswith('__'):
                    continue

                col = getattr(DefaultColsMixin, field)

                # Only consider SQLAlchemy columns
                if isinstance(col, sa.sql.schema.Column):
                    inherited_columns.append(field)

        combined_columns = set(constraint_columns + inherited_columns)
        if len(combined_columns) != len(expected_columns):

            missing = set(expected_columns) - combined_columns

            raise AssertionError(
                'Missing {} constraint tests for {}.'.format(
                    self.entity_cls.__name__,
                    ', '.join(missing))
            )

    def check_unique_constraint(self, **kwargs):
        self.entity_cls.testing_create(**kwargs)
        try:
            self.entity_cls.testing_create(**kwargs)
            raise AssertionError('Uniqueness error was not encountered.')
        except Exception as e:
            if not validate_unique_exc(e):
                raise


class SentryCapture:
    def __enter__(self):
        SentryClient.__log_reports__ = True
        SentryClient.__report_log__ = []
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        SentryClient.__log_reports__ = False

    @property
    def reports(self):
        return SentryClient.__report_log__



import sys
from decimal import Decimal

import _pytest
import pytest
import validators
import keg
import sqlalchemy as sa

import keg_elements.db.utils as dbutils
import kegel_app.model.entities as ents


def test_randemail():
    for length in range(0, 6):
        with pytest.raises(ValueError):
            dbutils.randemail(length)

    def not_so_rand_email(length, char):
        return dbutils.randemail(length, randomizer=lambda n, *args, **kwargs: char * n)

    assert not_so_rand_email(7, 'a') == 'a@a.aaa'
    assert not_so_rand_email(8, 'b') == 'b@bb.bbb'
    assert not_so_rand_email(9, 'c') == 'cc@cc.ccc'

    # Fuzzy testing for an extra dose of confidence
    for length in range(7, 50):
        email = dbutils.randemail(length)
        assert len(email) == length
        assert validators.email(email)

    # Get some confidence of randomness
    assert len(set(dbutils.randemail(50) for _ in range(5))) == 5, \
        'randemail not random (beware non-determinism; try again)'


class TestUpdateCollection(object):

    def setup_method(self, method):
        keg.db.db.session.remove()
        ents.Thing.delete_cascaded()
        ents.RelatedThing.delete_cascaded()

    def test_add(self):
        thing = ents.Thing.testing_create()
        assert not thing.related_things

        data = [{'name': 'test', 'is_enabled': True}]
        thing.update_collection('related_things', data)

        assert len(thing.related_things) == 1
        related = thing.related_things[0]
        assert related.name == 'test'
        assert related.is_enabled

    def test_edit(self):
        thing = ents.Thing.testing_create()
        related1_id = ents.RelatedThing.testing_create(thing=thing, name='a').id
        related2_id = ents.RelatedThing.testing_create(thing=thing, name='x').id

        data = [
            {'id': related1_id, 'name': 'b', 'is_enabled': False},
            {'id': related2_id, 'name': 'y', 'is_enabled': False},
        ]
        thing.update_collection('related_things', data)

        keg.db.db.session.flush()

        assert len(thing.related_things) == 2
        related1 = thing.related_things[0]
        assert related1.id == related1_id
        assert related1.name == 'b'
        assert not related1.is_enabled

        related2 = thing.related_things[1]
        assert related2.id == related2_id
        assert related2.name == 'y'
        assert not related2.is_enabled

    def test_append_id_none(self):
        thing = ents.Thing.testing_create()
        related = ents.RelatedThing.testing_create(thing=thing, name='a',
                                                   is_enabled=True)

        data = [
            {'is_enabled': False}
        ]
        thing.update_collection('related_things', data)

        assert len(thing.related_things) == 1
        related = thing.related_things[0]

        assert related.name is None
        assert not related.is_enabled

    def test_remove(self):
        thing = ents.Thing.testing_create()
        ents.RelatedThing.testing_create(thing=thing, name='a', is_enabled=True)
        assert len(thing.related_things) == 1

        data = []
        thing.update_collection('related_things', data)

        assert len(thing.related_things) == 0


class TestExceptionHelpers:
    def setup_method(self, _):
        ents.ConstraintTester.delete_cascaded()

    def test_validate_unique_exception_unnamed(self):
        ents.ConstraintTester.testing_create(unique1=1)
        with pytest.raises(Exception) as exc:
            ents.ConstraintTester.testing_create(unique1=1)
        assert dbutils.validate_unique_exc(exc.value) is True

        with pytest.raises(Exception) as exc:
            ents.ConstraintTester.testing_create(check=101)
        assert dbutils.validate_unique_exc(exc.value) is False

    def test_validate_unique_exception_named(self):
        ents.ConstraintTester.testing_create(unique2=1)
        with pytest.raises(Exception) as exc:
            ents.ConstraintTester.testing_create(unique2=1)
        assert dbutils.validate_unique_exc(exc.value, 'uq_constraint_tester_unique2') is True

    def test_validate_unique_msg_postgres(self):
        postgres_msg = (
            '(psycopg2.errors.UniqueViolation) duplicate key value violates unique constraint '
            '"uq_constraint_tester_unique2"'
        )
        assert dbutils._validate_unique_msg('postgresql', postgres_msg) is True
        assert dbutils._validate_unique_msg(
            'postgresql',
            postgres_msg,
            'uq_constraint_tester_unique2'
        ) is True

        assert dbutils._validate_unique_msg(
            'postgresql',
            postgres_msg,
            'uq_constraint_tester_unique1'
        ) is False
        assert dbutils._validate_unique_msg('mssql', postgres_msg) is False

    def test_validate_unique_msg_mssql(self):
        mssql_msg = (
            "Violation of UNIQUE constraint 'uq_constraint_tester_unique2'. "
            "Cannot insert duplicate key in object 'dbo.constraint_tester'. "
            'The duplicate key value is (1).'
        )
        assert dbutils._validate_unique_msg('mssql', mssql_msg) is True
        assert dbutils._validate_unique_msg(
            'mssql',
            mssql_msg,
            'uq_constraint_tester_unique2'
        ) is True

        assert dbutils._validate_unique_msg(
            'mssql',
            mssql_msg,
            'uq_constraint_tester_unique1'
        ) is False
        assert dbutils._validate_unique_msg('postgresql', mssql_msg) is False

    def test_validate_unique_msg_sqlite(self):
        sqlite_msg = '(sqlite3.IntegrityError) UNIQUE constraint failed: constraint_tester.unique2'
        assert dbutils._validate_unique_msg('sqlite', sqlite_msg) is True
        assert dbutils._validate_unique_msg(
            'sqlite',
            sqlite_msg,
            'uq_constraint_tester_unique2'
        ) is True

        assert dbutils._validate_unique_msg('postgresql', sqlite_msg) is False

    def test_validate_unique_msg_unsupported(self):
        sqlite_msg = '(sqlite3.IntegrityError) UNIQUE constraint failed: constraint_tester.unique2'
        with pytest.raises(ValueError) as exc:
            dbutils._validate_unique_msg('unknown', sqlite_msg)
        assert str(exc.value) == 'is_unique_exc() does not yet support dialect: unknown'

    @pytest.mark.skipif(sys.version_info.major < 3, reason='requires python 3')
    def test_raises_unique_exception(self):
        ents.ConstraintTester.testing_create(unique2=1)

        @dbutils.raises_unique_exc('uq_constraint_tester_unique2')
        def make_dup():
            ents.ConstraintTester.testing_create(unique2=1)

        make_dup()

        @dbutils.raises_unique_exc('uq_constraint_tester_unique2')
        def make_uq():
            ents.ConstraintTester.testing_create()

        with pytest.raises(_pytest.outcomes.Failed):
            make_uq()

        @dbutils.raises_unique_exc('uq_constraint_tester_unique2')
        def make_ck():
            ents.ConstraintTester.testing_create(check=101)

        with pytest.raises(AssertionError):
            make_ck()

    def test_is_check_constraint_sqlite(self):
        assert dbutils._is_check_const(
            'sqlite',
            '(sqlite3.IntegrityError) CHECK constraint failed: ck_constraint_tester_check',
            'ck_constraint_tester_check'
        ) is True

        assert dbutils._is_check_const(
            'sqlite',
            '(sqlite3.IntegrityError) UNIQUE constraint failed: constraint_tester.unique2',
            'ck_constraint_tester_check'
        ) is False

    def test_is_check_constraint_mssql(self):
        mssql_msg = (
            'The INSERT statement conflicted with the CHECK constraint '
            '"ck_constraint_tester_check".'
        )
        assert dbutils._is_check_const('mssql', mssql_msg, 'ck_constraint_tester_check') is True

        assert dbutils._is_check_const('mssql', mssql_msg, 'uq_constraint_tester_unique2') is False
        assert dbutils._is_check_const('sqlite', mssql_msg, 'ck_constraint_tester_check') is False

    def test_is_check_constraint_postgres(self):
        postgres_msg = (
            '(psycopg2.errors.CheckViolation) new row for relation "constraint_tester" violates '
            'check constraint "ck_constraint_tester_check"'
        )
        assert dbutils._is_check_const(
            'postgresql',
            postgres_msg,
            'ck_constraint_tester_check'
        ) is True

        assert dbutils._is_check_const(
            'postgresql',
            postgres_msg,
            'uq_constraint_tester_unique2'
        ) is False
        assert dbutils._is_check_const(
            'sqlite',
            postgres_msg,
            'ck_constraint_tester_check'
        ) is False

    def test_is_check_constraint_unsupported(self):
        sqlite_msg = '(sqlite3.IntegrityError) CHECK constraint failed: ck_constraint_tester_check'
        with pytest.raises(ValueError) as exc:
            dbutils._is_check_const('unknown', sqlite_msg, 'ck_constraint_tester_check')
        assert str(exc.value) == 'is_constraint_exc() does not yet support dialect: unknown'

    @pytest.mark.skipif(sys.version_info.major < 3, reason='requires python 3')
    def test_raises_check_exc(self):
        @dbutils.raises_check_exc('ck_constraint_tester_check')
        def make_fail():
            ents.ConstraintTester.testing_create(check=101)

        make_fail()

        @dbutils.raises_check_exc('ck_constraint_tester_check')
        def make_pass():
            ents.ConstraintTester.testing_create()

        with pytest.raises(_pytest.outcomes.Failed):
            make_pass()

        @dbutils.raises_check_exc('ck_constraint_tester_check')
        def make_uq():
            ents.ConstraintTester.testing_create(unique1=1)
            ents.ConstraintTester.testing_create(unique1=1)

        with pytest.raises(AssertionError):
            make_uq()


# Repeat parameter used to somewhat counteract the non-deterministic nature of the tested function
@pytest.mark.parametrize('repeat', range(1, 20))
@pytest.mark.parametrize('scale,prec,max', [
    (1, 1, Decimal('0.9')),
    (2, 1, Decimal('9.9')),
    (3, 3, Decimal('0.999')),
    (3, 2, Decimal('9.99')),
    (3, 1, Decimal('9.9')),
    (3, 0, Decimal('99.0')),
    (4, 4, Decimal('0.9999')),
    (4, 3, Decimal('9.999')),
    (4, 2, Decimal('99.99')),
    (4, 1, Decimal('99.9')),
    (4, 0, Decimal('99.0')),
])
def test_random_numeric(scale, prec, max, repeat):
    col = sa.Column(sa.Numeric(scale, prec))
    value = dbutils.random_numeric(col)
    assert value <= max
    assert value >= -max
